# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Base classes for widget plugins.
"""

import os
from os import path
import sys
from textwrap import dedent
import re
import threading

from genshi.template import MarkupTemplate
from genshi.filters import HTMLFormFiller

from compat import json
from freezable import Freezable, freeze
import errors
from state import Statefull, Static, SetOnce, Dynamic, StaticProp, DynamicProp
from constants import TYPE_ERROR, OVERRIDE_ERROR


HOOKS = (
    '_finalize_parent',
    '_finalize_children',
    '_finalize_final',
)


def relurl(dst, src):
    if not (src and dst):
        return dst
    if dst.startswith('/') or src.startswith('/'):
        return dst
    (dst, src) = (filter(None, u.split('/')) for u in (dst, src))
    n = min(len(u) for u in (dst, src))
    for i in xrange(n):
        (d, s) = (dst[i], src[i])
        if d != s:
            break
    url = ['..' for p in src[i:]]
    url.extend(dst[i:])
    return '/'.join(url)


class Widget(Statefull):
    """
    Base class from which all widget plugins subclass.
    """

    # The genshi template:
    xml = None
    xml_t = None
    __tag__ = None

    # CSS style information (via style.Style):
    style_vars = None
    style_global = None
    style = None

    # JavaScript:
    js_class = None
    javascript = None

    # Plugable asset registration
    register_assets = None  # classmethod register_assets(cls, assets)
    require_js = tuple()
    require_css = tuple()

    # How id is constructed from an ancestor:
    base_id_from = 'parent'

    # Override this with an iterable providing extra CSS classes to be added to
    # css_classes @DynamicProp
    extra_css_classes = tuple()


    # -- Static --
    name = Static('name', doc="A widget name like 'my_menu'")
    id = Static('id', doc="A widget id like 'my_menuset.my_menu'")
    app = Static('app', doc='The `Application` this widget was created via')


    # -- StaticProp --
    @StaticProp
    def klass(self):
        return self.__class__.__name__

    @StaticProp
    def tag(self):
        return self.__class__.__tag__

    # -- DynamicProp --
    @DynamicProp
    def css_classes(self):
        extra = self.extra_css_classes
        if not extra:
            return self.klass
        classes = set([self.klass])
        classes.update(extra)
        return ' '.join(sorted(classes))

    __compiled = dict()
    __pat = re.compile(r'\W*<(\w+)\W')

    @classmethod
    def _compile(cls):
        if isinstance(cls.xml, basestring):
            if cls.xml not in cls.__compiled:
                cls.xml = dedent(cls.xml).strip() + '\n'
            if cls.xml not in cls.__compiled:
                cls.xml_t = MarkupTemplate(cls.xml)
                cls.__tag__ = cls.__pat.match(cls.xml).group(1)
                cls.__compiled[cls.xml] = (cls.xml_t, cls.__tag__)
            else:
                (xml_t, tag) = cls.__compiled[cls.xml]
                if cls.xml_t is not xml_t:
                    cls.xml_t = xml_t
                if cls.__tag__ != tag:
                    cls.__tag__ = tag

        if isinstance(cls.javascript, basestring):
            cls.javascript = dedent(cls.javascript).strip() + '\n'

    def __repr__(self):
        return '%s(id=%r, name=%r)' % (self.klass, self.id, self.name)

    def __json__(self):
        return self.id

    def finalize(self):
        if self.id is None and self.name is not None:
            if (
                self.base_id_from
                and getattr(self, self.base_id_from) is not None
            ):
                base = getattr(self, self.base_id_from).id
            else:
                base = None
            if base:
                self.id = '.'.join((base, self.name))
            else:
                self.id = self.name
        for name in HOOKS:
            hook = getattr(self, name, None)
            if callable(hook):
                hook()
        freeze(self)

    def _relurl(self, target):
        """
        Create a relative URL.
        """
        if isinstance(self, Page):
            base = self.urldir
        elif self.page is not None:
            base = self.page.urldir
        else:
            return target
        return relurl(target, base)
    relurl = StaticProp(name='relurl')

    def new(self, klass, **kw):
        if self.app is None:
            raise errors.MissingApplication(klass=klass, widget=self)
        return self.app.new(klass, **kw)

    def print_tree(self, fileobj=sys.stdout, level=0):
        indent = ' ' * (level * 4)
        fileobj.write(indent + repr(self) + '\n')

    def generate(self, **extra):
        if not self.xml_t:
            return
        state = self.state()
        if extra:
            state.update(extra)
        return self.xml_t.generate(**state)

    def render(self, **extra):
        if not self.xml_t:
            return ''
        return self.generate(**extra).render('xhtml')


class Anon(object):
    def __init__(self):
        self.__d = dict()

    def next(self, child):
        i = self.__d.get(child.klass, 1)
        self.__d[child.klass] = i + 1
        return '%s%d' % (child.klass.lower(), i)


class Container(Widget):
    '''
    A widget that can contain other widgets.

    For example, a basic Genshi template for a `Container` widget would look
    something like this:

    >>> class Parent(Container):
    ...     xml = """
    ...     <div
    ...         xmlns:py="http://genshi.edgewall.org/"
    ...         id="${id}"
    ...         class="${css_classes}"
    ...     >
    ...     <span py:for="child in children" py:replace="child.generate()" />
    ...     </div>
    ...     """
    ...
    >>> foo = Parent(id='foo')
    >>> print foo.render()
    <div id="foo" class="Parent">
    </div>

    Notice that above the ``<span>`` tag isn't rendered at all as the
    `Container` has no children at this point.

    Children can be incrementally added using a simple container API similar to
    a traditional toolkit like GTK+ or Qt.  For example, lets first define a
    simple widget class from which we'll create the children, and then create
    two ``Child`` instances:

    >>> class Child(Widget):
    ...     xml = """
    ...     <p
    ...         xmlns:py="http://genshi.edgewall.org/"
    ...         id="${id}"
    ...         class="${css_classes}"
    ...     />
    ...     """
    ...
    >>> bar = Child(name='bar')
    >>> baz = Child(name='baz')

    And now we'll add ``Child`` instances to ``foo`` using `Container.add()`:

    >>> foo.children
    ()
    >>> foo.add(bar, baz)
    >>> foo.children
    (Child(id=None, name='bar'), Child(id=None, name='baz'))
    >>> bar.parent is foo
    True
    >>> baz.parent is foo
    True

    Notice that when we render ``foo`` now, the ``<span>`` tag still isn't
    rendered because of the Genshi ``py:replace`` directive:

    >>> print foo.render()
    <div id="foo" class="Parent">
    <p class="Child"></p><p class="Child"></p>
    </div>

    The `Container.print_tree()` method can be used to easily visualize an
    entire parent/child hierarchy.  For example:

    >>> import sys
    >>> foo.print_tree(sys.stdout)
    Parent(id='foo', name=None)
        Child(id=None, name='bar')
        Child(id=None, name='baz')

    Or after we add a `Container` that itself has children:

    >>> sub = Parent()  # Lets add more children to foo
    >>> sub.add(Child(), Child(), Child())
    >>> foo.add(sub)
    >>> foo.print_tree(sys.stdout)
    Parent(id='foo', name=None)
        Child(id=None, name='bar')
        Child(id=None, name='baz')
        Parent(id=None, name=None)
            Child(id=None, name=None)
            Child(id=None, name=None)
            Child(id=None, name=None)

    Calling `Container.finalize()` will prepare the `Container` for serving
    or rendering to static files.  Among other things, it preforms the stable
    anonymous name and ID assignment.  For example:

    >>> foo.finalize()
    >>> foo.print_tree(sys.stdout)
    Parent(id='foo', name=None)
        Child(id='foo.bar', name='bar')
        Child(id='foo.baz', name='baz')
        Parent(id='foo.parent1', name='parent1')
            Child(id='foo.parent1.child1', name='child1')
            Child(id='foo.parent1.child2', name='child2')
            Child(id='foo.parent1.child3', name='child3')

    A few examples will best explain the why the anonymous name and ID
    assignment is said to be *stable*.  First, lets consider an example
    containing only anonymous children:

    >>> eg1 = Parent(id='foo')
    >>> eg1.add(
    ...     Child(), Child(), Child()
    ... )
    >>> eg1.finalize()
    >>> eg1.print_tree(sys.stdout)
    Parent(id='foo', name=None)
        Child(id='foo.child1', name='child1')
        Child(id='foo.child2', name='child2')
        Child(id='foo.child3', name='child3')

    Compared to an example where we explicitly assign a name to the second
    child:

    >>> eg2 = Parent(id='foo')
    >>> eg2.add(
    ...     Child(), Child(name='bar'), Child()
    ... )
    >>> eg2.finalize()
    >>> eg2.print_tree(sys.stdout)
    Parent(id='foo', name=None)
        Child(id='foo.child1', name='child1')
        Child(id='foo.bar', name='bar')
        Child(id='foo.child3', name='child3')

    Compared to an example where we explicitly assign an ID to the second child:

    >>> eg3 = Parent(id='foo')
    >>> eg3.add(
    ...     Child(), Child(id='bar'), Child()
    ... )
    >>> eg3.finalize()
    >>> eg3.print_tree(sys.stdout)
    Parent(id='foo', name=None)
        Child(id='foo.child1', name='child1')
        Child(id='bar', name='child2')
        Child(id='foo.child3', name='child3')

    Compared to an example where we insert something besides a ``Child`` widget
    between the first and second ``Child`` instances:

    >>> eg4 = Parent(id='foo')
    >>> eg4.add(
    ...     Child(), Parent(), Child(), Child()
    ... )
    >>> eg4.finalize()
    >>> eg4.print_tree(sys.stdout)
    Parent(id='foo', name=None)
        Child(id='foo.child1', name='child1')
        Parent(id='foo.parent1', name='parent1')
        Child(id='foo.child2', name='child2')
        Child(id='foo.child3', name='child3')

    The idea with the stable anonymous assignment is that it allows you to make
    fairly dramatic changes to your `Page` tree while changing as few
    anonymously assigned names and IDs as possible.

    `Container.finalize()` does not perform anonymous name or ID assignment for
    children that subclass from `Form` or `Field`.   For example:

    >>> eg = Parent(id='foo')
    >>> eg.add(
    ...     Form(), Field()
    ... )
    >>> eg.finalize()
    >>> eg.print_tree(sys.stdout)
    Parent(id='foo', name=None)
        Form(id=None)
        Field(id=None, name=None)

    Anonymous `Form` IDs are assigned relative to their `Page` when
    `Page.finalize()` is called.  Anonymous `Field` names are assigned relative
    to their `Form` when `Form.finalize()` is called.
    '''

    def children(self):
        return tuple(self.__children)

    children = StaticProp(children, json=True)

    xml = """
    <div
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
    >
    <child py:for="child in children" py:replace="child.generate()" />
    </div>
    """

    def __init__(self, **kw):
        self.__neutered = False
        self.__children = []
        self.__all_children = None
        self.__names = dict()
        super(Container, self).__init__(**kw)
        self.populate()

    def populate(self):
        pass

    def __len__(self):
        """
        Return the number of children in this container.
        """
        return len(self.__children)

    def iter(self):
        """
        Iterate through the children (direct descendants) in this container.
        """
        for child in self.__children:
            yield child

    def deepiter(self):
        """
        Iterate recursively through all child widgets in this container.
        """
        if self.__all_children is None:
            for child in self.__children:
                yield child
                if isinstance(child, Container):
                    for c in child.deepiter():
                        yield c
        else:
            for child in self.__all_children:
                yield child

    def print_tree(self, fileobj=sys.stdout, level=0):
        super(Container, self).print_tree(fileobj, level)
        for child in self.iter():
            child.print_tree(fileobj, level + 1)

    def neuter(self, check=True):
        """
        A neutered `Container` can't have any more children.
        """
        if self.__neutered:
            if check:
                raise errors.AlreadyCalled(
                    klass=self.klass, method='neuter', inst=self
                )
            return False
        self.__neutered = True
        self.__children = tuple(self.__children)
        return True

    def deepneuter(self, check=False):
        """
        Neuter this and all descendant `Container`.
        """
        self.neuter(check)
        for child in self.__children:
            if isinstance(child, Container):
                child.deepneuter()
        self.__all_children = tuple(self.deepiter())

    def add(self, *children):
        """
        Add one or more child widgets to this container.

        For example:

        >>> parent = Container()
        >>> child1 = Widget()
        >>> child2 = Widget()
        >>> parent.add(child1, child2)
        >>> list(parent.iter()) == [child1, child2]
        True
        >>> child1.parent is parent
        True
        >>> child2.parent is parent
        True
        """
        if self.__neutered:
            raise errors.NeuteredContainer(children=children, container=self)
        for child in children:
            if not isinstance(child, Widget):
                raise TypeError(
                    TYPE_ERROR % ('child', Widget, type(child), child)
                )
            if isinstance(child, Page):
                raise TypeError(
                    'child: a %r instance cannot be a child' % Page
                )
            child.parent = self
            assert child not in self.__children
            self.__children.append(child)

    def new_child(self, klass, **kw):
        child = self.new(klass, **kw)
        self.add(child)
        return child

    def _finalize_parent(self):
        anon = Anon()
        for child in self.iter():
            if isinstance(child, (Form, Field)):
                continue
            anon_name = anon.next(child)
            if child.name is None:
                child.name = anon_name
            if child.name in self.__names:
                raise errors.NameConflict(
                    new=child, orig=self.__names[child.name], parent=self
                )
            self.__names[child.name] = child

    def _finalize_children(self):
        for child in self.children:
            child.finalize()


Widget.parent = SetOnce('parent', kind=Container, json=True)


class Form(Container):
    """
    Base class for all HTML form element widgets.

    A `Form` is much like a regular `Container` except that any ancestors which
    are `Field` instances are treated specially.

    `Form` instances provide some basic Python container emulation mapping field
    names to fields, although the mapping is not initialized till
    `Form.finalize()` is called.  For example, if we populate a `Form` like
    this:

    >>> form = Form(id='foo')
    >>> container = Container(name='bar')
    >>> form.add(Field(), Widget(), container, Field(), Widget())
    >>> container.add(Field(), Widget())

    Initially no mapping is available:

    >>> list(form)
    []
    >>> 'field2' in form
    False
    >>> form['field2']
    Traceback (most recent call last):
      ...
    KeyError: 'field2'

    But the mapping will be initialized after we call `Form.finalize()`:

    >>> form.finalize()
    >>> list(form)
    ['field1', 'field2', 'field3']
    >>> 'field2' in form
    True
    >>> form['field2']
    Field(id='foo.field2', name='field2')

    Anonymous names for `Field` instances are assigned relative to the `Form`
    containing them, instead of relative to the immediate parent, which might
    not be the containing `Form`, as you can see here:

    >>> import sys
    >>> form.print_tree(sys.stdout)
    Form(id='foo')
        Field(id='foo.field1', name='field1')
        Widget(id='foo.widget1', name='widget1')
        Container(id='foo.bar', name='bar')
            Field(id='foo.field2', name='field2')
            Widget(id='foo.bar.widget1', name='widget1')
        Field(id='foo.field3', name='field3')
        Widget(id='foo.widget2', name='widget2')

    `Field` names must be unique in the namespace of the `Form`, or otherwise
    `errors.FieldConflict` is raised.  For example:

    >>> form = Form(id='foo')
    >>> container = Container()
    >>> form.add(Field(name='bar', id='first'), container)
    >>> container.add(Field(name='bar'))
    >>> form.finalize()
    Traceback (most recent call last):
      ...
    FieldConflict: Field(id=None, name='bar') `name` conflicts with Field(id='first', name='bar') in Form(id='foo')

    For ancestors that are not `Field` instances, a `Form` behaves exactly like
    a `Container`.  See `Container` for additional details.
    """

    base_id_from = None
    name = None  # Form does not have @Static `name`
    action = Dynamic('action')
    method = Dynamic('method', default='POST')

    xml = """
    <form
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
        method="${method}"
        action="${relurl(action)}"
    >
    <input py:for="child in children" py:replace="child.generate()" />
    </form>
    """

    def __init__(self, **kw):
        super(Form, self).__init__(**kw)
        self.__fields = dict()

    def __repr__(self):
        return '%s(id=%r)' % (self.klass, self.id)

    def __iter__(self):
        """
        Iterate in ascending order through the names of fields in this `Form`.

        The name-to-field map isn't built till `Form.finalize()` is called,
        before which this method will always return an empty generator.

        For example:

        >>> form = Form()
        >>> container = Container()
        >>> form.add(Field(name='foo'), container)
        >>> container.add(Field(name='bar'), Field(name='baz'))
        >>> list(form)
        []
        >>> form.finalize()
        >>> list(form)
        ['bar', 'baz', 'foo']
        """
        for name in sorted(self.__fields):
            yield name

    def __contains__(self, name):
        """
        Return ``True`` if `Form` contains a `Field` named *name*.

        The name-to-field map isn't built till `Form.finalize()` is called,
        before which this method will always return ``False``.

        For example:

        >>> form = Form()
        >>> container = Container()
        >>> form.add(container)
        >>> container.add(Field(name='email'))
        >>> 'email' in form
        False
        >>> form.finalize()
        >>> 'email' in form
        True

        :param name: The name of the field in question, e.g. ``'email'``
        """
        return name in self.__fields

    def __getitem__(self, name):
        """
        Return the `Field` instance named *name*.

        The name-to-field map isn't built till `Form.finalize()` is called,
        before which this method will always raise a ``KeyError``.

        For example:

        >>> form = Form(id='foo')
        >>> container = Container()
        >>> form.add(container)
        >>> container.add(Field(name='bar'))
        >>> form['bar']
        Traceback (most recent call last):
          ...
        KeyError: 'bar'
        >>> form.finalize()
        >>> form['bar']
        Field(id='foo.bar', name='bar')

        After `Form.finalize()` has been called, this method only raises a
        ``KeyError`` if no such field exists.  For example:

        >>> form['baz']
        Traceback (most recent call last):
          ...
        KeyError: 'baz'

        :param name: The name of the field in question, e.g. ``'email'``
        """
        return self.__fields[name]

    def _finalize_parent(self):
        super(Form, self)._finalize_parent()
        anon = Anon()
        for child in self.deepiter():
            if not isinstance(child, Field):
                continue
            anon_name = anon.next(child)
            if child.name is None:
                child.name = anon_name
            if child.name in self.__fields:
                raise errors.FieldConflict(
                    new=child, orig=self.__fields[child.name], form=self
                )
            self.__fields[child.name] = child
            child.form = self

    def generate(self, **extra):
        stream = super(Form, self).generate(**extra)
        if self.__fields:
            data = dict(
                (k, f.value) for (k, f) in self.__fields.iteritems()
            )
            return stream.filter(HTMLFormFiller(data=data))
        return stream


class Field(Widget):
    base_id_from = 'form'
    form = SetOnce('form', kind=Form)

    label = Dynamic('label', default='Field Label')
    required = Dynamic('required', default=False)
    value = Dynamic('value')
    error = Dynamic('error')


class Page(Container):
    """
    Top level container representing a page.

    A page cannot be a child in another widget.
    """

    # Overrides
    name = None  # Page does not have @Static `name`
    parent = None  # Page does not have @SetOnce `parent`
    page = None  # Page does not have @SetOnce `page`

    # -- Static --
    inline_js = Static('inline_js')
    inline_css = Static('inline_css')

    id = Static('id', default='',
        doc="A page ID like 'reports:daily:today'"
    )

    static = Static('static', default=False,
        doc='If True, page is configured for rendering to static files'
    )

    def _validate_baseurl(self, value):
        if value:
            return value.rstrip('/') + '/'
        return ''
    baseurl = Static('baseurl', validator=_validate_baseurl, default='')

    def _validate_serializer(self, value):
        t = ('xhtml', 'html')
        if value not in t:
            raise ValueError('%s.serializer must be in %r; got %r' % (
                self.klass, t, value)
            )
        return value
    serializer = Static('serializer',
        validator=_validate_serializer,
        default='xhtml',
        doc='The Genshi serializer to use'
    )


    # -- Dynamic --
    title = Dynamic('title', default='Page Title')
    lang = Dynamic('lang', default='en')


    # -- StaticProp --
    @StaticProp
    def content_type(self):
        d = dict(
            xhtml='application/xhtml+xml; charset=utf-8',
            html='text/html; charset=utf-8',
        )
        return d[self.serializer]

    @StaticProp
    def doctype(self):
        d = dict(
            xhtml='xhtml11',
            html='html-strict',
        )
        return d[self.serializer]

    @StaticProp
    def urldir(self):
        """
        The "directory" portion of the URL, e.g. ``'/my-app/foo'``.
        """
        rel = '/'.join(self.get_url_parts())
        if self.baseurl.startswith('/'):
            return '/' + rel
        return rel

    @StaticProp
    def url(self):
        if not (self.static or self.id):
            return self.baseurl
        rel = '/'.join(self.get_url_parts(self.static))
        if self.baseurl.startswith('/'):
            return '/' + rel
        return rel


    # -- DynamicProp --
    @DynamicProp
    def links_css(self):
        return tuple(
            self._iter_links(self.assets_css)
        )

    @DynamicProp
    def links_js(self):
        return tuple(
            self._iter_links(self.assets_js)
        )

    def _iter_links(self, keys):
        if self.app is None:
            return
        for key in keys:
            try:
                yield self._relurl(self.app.assets.asset_url(key))
            except KeyError:
                pass

    xml = """
    <html
        xmlns:py="http://genshi.edgewall.org/"
        xmlns="http://www.w3.org/1999/xhtml"
        xml:lang="${lang}"
    >
    <head>
    <meta http-equiv="Content-Type" content="${content_type}" />
    <title py:content="title" />
    <link
        py:for="link in links_css"
        href="${link}"
        rel="stylesheet"
        type="text/css"
    />
    <style
        py:if="inline_css"
        py:content="inline_css"
        type="text/css"
    />
    <script
        py:for="link in links_js"
        src="${link}"
        type="text/javascript"
    />
    <script
        py:if="inline_js"
        py:content="inline_js"
        type="text/javascript"
    />
    </head>
    <body class="${css_classes}" id="${id}">
    <div
        py:for="child in children"
        py:if="child"
        py:replace="child.generate()"
    />
    </body>
    </html>
    """

    assets_js = (
        'wehjit.js',
    )
    assets_css = (
        'wehjit.css',
    )

    controller = None

    def __init__(self, **kw):
        super(Page, self).__init__(**kw)
        self.__ids = dict()
        self.__plugins = None

    def __repr__(self):
        return '%s(id=%r)' % (self.klass, self.id)

    def __iter__(self):
        """
        Iterate in ascending order through the IDs of widgets in this `Page`.

        The ID-to-child map isn't built till `Page.finalize()` is called, before
        which this method will always return an empty generator.

        For example:

        >>> page = Page()
        >>> container = Container(id='hello')
        >>> widget = Widget(id='world')
        >>> page.add(container)
        >>> container.add(widget)
        >>> list(page)
        []
        >>> page.finalize()
        >>> list(page)
        ['hello', 'world']
        """
        for id in sorted(self.__ids):
            yield id

    def __contains__(self, id):
        """
        Return ``True`` if `Page` contains a child with ID *id*.

        The ID-to-child map isn't built till `Page.finalize()` is called, before
        which this method will always return ``False``.

        For example:

        >>> page = Page()
        >>> container = Container(id='foo')
        >>> widget = Widget(name='bar')
        >>> page.add(container)
        >>> container.add(widget)
        >>> 'foo' in page
        False
        >>> 'foo.bar' in page
        False
        >>> page.finalize()
        >>> 'foo' in page
        True
        >>> 'foo.bar' in page
        True

        :param id: The ID of the child in question, e.g. ``'menuset1.menu2'``
        """
        return id in self.__ids

    def __getitem__(self, id):
        """
        Return the child with ID *id*.

        The ID-to-child map isn't built till `Page.finalize()` is called, before
        which this method will always raise a ``KeyError``.

        For example:

        >>> page = Page()
        >>> container = Container(id='foo')
        >>> widget = Widget(name='bar')
        >>> page.add(container)
        >>> container.add(widget)
        >>> page['foo.bar']
        Traceback (most recent call last):
          ...
        KeyError: 'foo.bar'
        >>> page.finalize()
        >>> page['foo.bar']
        Widget(id='foo.bar', name='bar')

        After `Page.finalize()` has been called, this method only raises a
        ``KeyError`` if no such child exists.  For example:

        >>> page['bar:baz']
        Traceback (most recent call last):
          ...
        KeyError: 'bar:baz'

        Note that this method can retrieve any ancestor widget by ID, is not
        limited to only direct children of the `Page` itself.

        :param id: The ID of the child in question, e.g. ``'menuset1.menu2'``
        """
        return self.__ids[id]

    def plugins(self):
        """
        Iterate through the widget plugins that have been used at least once.

        The first item will always be the class of the `Page` plugin itself,
        followed by the widget plugins (Python classes) of all ancestors,
        sorted by class name.

        For example:

        >>> page = Page()
        >>> list(page.plugins())
        [<class 'wehjit.base.Page'>]
        >>> c1 = Container()
        >>> w1 = Widget()
        >>> page.add(c1)
        >>> c1.add(w1)
        >>> list(page.plugins())
        [<class 'wehjit.base.Page'>, <class 'wehjit.base.Container'>, <class 'wehjit.base.Widget'>]
        >>> c2 = Container()
        >>> w2 = Widget()
        >>> page.add(c2, w2)

        >>> list(page.plugins())
        [<class 'wehjit.base.Page'>, <class 'wehjit.base.Container'>, <class 'wehjit.base.Widget'>]
        """
        if self.__plugins is None:
            plugins =  self._get_plugins()
        else:
            plugins = self.__plugins
        for p in plugins:
            yield p

    def _get_plugins(self):
        plugins = set(child.__class__ for child in self.deepiter())
        return [self.__class__] + sorted(plugins, key=lambda p: p.__name__)

    def finalize(self):
        """
        Finalize all children and put `Page` into a frozen, thread-safe state.
        """
        self.deepneuter()
        super(Page, self).finalize()

    def _finalize_parent(self):
        super(Page, self)._finalize_parent()
        anon = Anon()
        for child in self.deepiter():
            if isinstance(child, Form):
                anon_name = anon.next(child)
                if child.name is None:
                    child.name = anon_name
            child.page = self

    def _finalize_final(self):
        for child in self.deepiter():
            if child.id in self.__ids:
                raise errors.IDConflict(
                    new=child, orig=self.__ids[child.id], page=self
                )
            self.__ids[child.id] = child
        self.__plugins = tuple(self._get_plugins())
        self._prepend_require('css')
        self._prepend_require('js')

    def _iter_js_classes(self):
        for c in self.plugins():
            if c.js_class:
                yield dict(
                    uses=c.js_class,
                    klass=c.__name__,
                    tag=c.__tag__,
                )

    def _iter_json(self):
        for c in self.deepiter():
            if c.JSONKeys:
                yield (c.id, dict((k, getattr(c, k)) for k in c.JSONKeys))

    def _prepend_require(self, ext):
        require = 'require_' + ext
        assets = 'assets_' + ext
        acum = set()
        for plugin in self.plugins():
            acum.update(
                getattr(plugin, require)
            )
        result = tuple(acum) + getattr(self, assets)
        setattr(self, assets, result)

    def render(self):
        return self.generate().render(self.serializer,
            doctype=self.doctype,
        )

    def get_url_parts(self, index=False):
        parts = self.baseurl.split('/')
        parts.extend(self.id.split(':'))
        if index:
            parts.append('index.' + self.serializer)
        return filter(None, parts)

    def filename(self, out):
        parts = self.get_url_parts(index=True)
        if self.baseurl.startswith('/'):
            prefix = 'abs'
        else:
            prefix = 'rel'
        parts.insert(0, prefix)
        return path.join(out, *parts)

    def render_static(self, out='static', reporter=None):
        f = self.filename(out)
        if reporter:
            reporter.step('Rendering page %r', self.id)
            reporter.info('url=%r', self.url)
            reporter.info('file=%r', f)
        if path.lexists(f):
            raise StandardError('static file already exists: %r' % f)
        d = path.dirname(f)
        if not path.lexists(d):
            os.makedirs(d)
        fp = open(f, 'w')
        fp.write(self.render())
        fp.close()

    def deepclear(self):
        """
        Recursively clear the thread-local request state.

        This will recursively clear the thread-local request state for this
        `Page` and for all ancestor widgets contained herein.
        """
        self.clear()
        for child in self.deepiter():
            child.clear()

    def dumps(self, indent=None):
        """
        Return the global request state in JSON encoding.

        For example:

        >>> from wehjit import builtins
        >>> page = builtins.Page(title='Hello')
        >>> h1 = builtins.H1()
        >>> page.add(h1)
        >>> page.finalize()
        >>> print page.dumps()
        {"": {"lang": "en", "title": "Hello"}, "h11": {"label": null}}
        >>> h1.label = 'world!'
        >>> print page.dumps()
        {"": {"lang": "en", "title": "Hello"}, "h11": {"label": "world!"}}
        """
        return json.dumps(self.deeprequest(), sort_keys=True, indent=indent)

    def deeprequest(self):
        """
        Return the total request state for this `Page` and all ancestors.

        >>> from wehjit import builtins
        >>> page = builtins.Page(title='Hello')
        >>> h1 = builtins.H1()
        >>> page.add(h1)
        >>> page.finalize()
        >>> page.deeprequest()
        {'': {'lang': 'en', 'title': 'Hello'}, 'h11': {'label': None}}
        >>> h1.label = 'world!'
        >>> page.deeprequest()
        {'': {'lang': 'en', 'title': 'Hello'}, 'h11': {'label': 'world!'}}

        Also see the `Statefull.request()` method.
        """
        return dict(self._deeprequest_iter())

    def _deeprequest_iter(self):
        yield ('', self.request())
        for child in self.deepiter():
                yield (child.id, child.request())

    controller = None

    def __call__(self, environ, start_response):
        """
        Simple WSGI application handler.
        """
        if callable(self.controller):
            self.deepclear()
            self.controller(environ)
        response = self.render()
        response_headers = [
            ('Content-Type', self.content_type),
        ]
        start_response('200 OK', response_headers)
        return [response]


Widget._page = None
Widget.page = SetOnce('page', kind=Page)
