# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Custom exceptions.
"""

class WehjitError(StandardError):
    """
    Base class for custom wehjit exceptions.
    """

    format = ''

    def __init__(self, **kw):
        for (key, value) in kw.iteritems():
            assert not hasattr(self, key), 'conflicting kwarg %s.%s = %r' % (
                self.__class__.__name__, key, value,
            )
            setattr(self, key, value)
        StandardError.__init__(self, self.format % kw)


class OverrideError(WehjitError):
    """
    Raised when a plugin overrides another without using ``override=True``.

    For example:

    >>> raise OverrideError(orig='Checkbox', new='<Plugin>')
    Traceback (most recent call last):
      ...
    OverrideError: unexpected override of 'Checkbox' with '<Plugin>'
    """

    format = 'unexpected override of %(orig)r with %(new)r'


class MissingOverrideError(WehjitError):
    """
    Raised when a plugin overrides another that has not been registered.

    For example:

    >>> raise MissingOverrideError(name='Checkbox', plugin='<Plugin>')
    Traceback (most recent call last):
      ...
    MissingOverrideError: Checkbox not registered, cannot override with '<Plugin>'
    """

    format = '%(name)s not registered, cannot override with %(plugin)r'


class AttributeConflict(WehjitError):
    """
    Raised when a plugin's name conflicts with an attribute on the registrar.

    For example:

    >>> raise AttributeConflict(plugin='<Plugin>', registrar='Registrar', attribute='register')
    Traceback (most recent call last):
      ...
    AttributeConflict: '<Plugin>' name conflicts with Registrar.register attribute
    """

    format = '%(plugin)r name conflicts with %(registrar)s.%(attribute)s attribute'


class NameConflict(WehjitError):
    """
    Raised when a child name conflicts with another in the same parent.
    """

    format = '%(new)r `name` conflicts with %(orig)r in %(parent)r'


class FieldConflict(WehjitError):
    """
    Raised when a field name conflicts with another in the same form.
    """

    format = '%(new)r `name` conflicts with %(orig)r in %(form)r'


class IDConflict(WehjitError):
    """
    Raised when a wehjit id conflicts with another in the same page.
    """

    format = '%(new)r `id` conflicts with %(orig)r in %(page)r'


class PageConflict(WehjitError):
    """
    Raised when a page id conflicts with another in the same application.
    """

    format = '%(new)r `id` conflicts with %(orig)r in %(app)r'


class AlreadyCalled(WehjitError):
    format = '%(klass)s.%(method)s() has already been called on %(inst)r'


class StyleVarError(WehjitError):
    """
    Raised when a CSS property references an undefined style-variable.

    For example:

    >>> raise StyleVarError(
    ...     var='b_color',
    ...     key='border',
    ...     value='1px solid %(b_color)s',
    ...     namespace='Menu',
    ...     selector='> ul',
    ... )
    Traceback (most recent call last):
      ...
    StyleVarError: undefined style var 'b_color'
        used in CSS prop 'border' = '1px solid %(b_color)s'
        in namespace 'Menu', selector '> ul'
    """

    format = """undefined style var %(var)r
    used in CSS prop %(key)r = %(value)r
    in namespace %(namespace)r, selector %(selector)r"""


class MissingApplication(WehjitError):
    format = 'new(%(klass)r): %(widget)r has no associated application'


class NeuteredContainer(WehjitError):
    format = 'cannot add %(children)r to neutered %(container)r'
