# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
`Field` widgets in grid layout inside a `FieldTable`.
"""

from wehjit import base, builtins
from wehjit.state import *


class FieldTable(base.Container):
    """
    Arrange `FieldRow` in a table.
    """

    xml = """
    <table
        xmlns:py="http://genshi.edgewall.org/"
        class="${klass}"
        id="${id}"
    >
    <tr py:for="child in children" py:replace="child.generate()" />
    </table>
    """

    style = (
        ('', (
            ('empty-cells', 'show'),
            ('border-collapse', 'collapse'),
            ('margin', '1em 2em'),
        )),

        ('tr.error', (
            ('background-color', '#ddd'),
        )),


        ('th', (
            ('text-align', 'right'),
            ('font-weight', 'normal'),
            ('padding', '0.25em 0.5em'),
            ('line-height', '%(height_bar)s'),
            ('vertical-align', 'top'),
        )),

        ('tr.required th', (
            ('font-weight', 'bold'),
        )),

        ('td', (
            ('padding', '0.25em 0.5em'),
            ('vertical-align', 'top'),
            ('text-align', 'left'),
        )),

        ('td.error', (
            ('color', '%(color_error)s'),
        ))
    )

builtins.register(FieldTable)


class FieldRow(base.Field):

    type = StaticProp(name='type')

    @property
    def extra_css_classes(self):
        if self.required:
            yield 'required'
        if self.error:
            yield 'error'

    xml = """
    <tr
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
    >
    <th><label for="${id}.__field__">${label}:</label></th>
    <td>
    <textarea
        py:if="type == 'textarea'"
        name="${name}"
        id="${id}.__field__"
        rows="${rows}"
        cols="${cols}"
        py:content="value"
    />
    <select
        py:if="type == 'select'"
        name="${name}"
        id="${id}.__field__"
    >
        <option
            py:for="(optvalue,optlabel) in options"
            py:content="optlabel"
            value="${optvalue}"
        />
    </select>
    <input
        py:if="type == 'text' or type == 'password'"
        name="${name}"
        id="${id}.__field__"
        type="${type}"
        size="${size}"
        maxlength="${maxlength}"
    />
    </td>
    <td class="error" py:content="error" />
    </tr>
    """


class TextRow(FieldRow):
    """
    A single-line text field to be placed in a `FieldTable`.
    """

    _type = 'text'

    size = Dynamic('size', default=30)
    maxlength = Dynamic('maxlength')

    style = (
        ('input', (
            ('padding', '%(padding_input)s'),
        )),
    )

builtins.register(TextRow)


class PasswordRow(TextRow):
    """
    A password and confirmation field to be placed in a `FieldTable`.
    """

    _type = 'password'
    _label = 'Password'
    confirm = Dynamic('confirm', default='Confirm Password')

    xml = """
    <tr
        xmlns:py="http://genshi.edgewall.org/"
        py:strip="True"
    >

    <tr
        class="${css_classes}"
        id="${id}"
    >
    <th><label for="${id}.__field__" py:content="label" /></th>
    <td>
    <input
        name="${name}"
        id="${id}.__field__"
        size="${size}"
        maxlength="${maxlength}"
        type="password"
    />
    </td>
    <td class="error" rowspan="2" py:content="error" />
    </tr>

    <tr class="${css_classes}">
    <th><label for="${id}.__confirm__" py:content="confirm" /></th>
    <td>
    <input
        name="${name}"
        id="${id}.__confirm__"
        size="${size}"
        maxlength="${maxlength}"
        type="password"
    />
    </td>
    </tr>

    </tr>
    """


builtins.register(PasswordRow)


class TextAreaRow(FieldRow):
    """
    A textarea field to be placed in a `FieldTable`.
    """

    _type = 'textarea'
    rows = Dynamic('rows', default=4)
    cols = Dynamic('cols', default=30)

    style = (
        ('textarea', (
            ('padding', '%(padding_input)s'),
        )),
    )

builtins.register(TextAreaRow)


class SelectRow(FieldRow):
    """
    A single dropdown to be placed in a `FieldTable`.
    """

    _type = 'select'
    options = Dynamic('options', default=(('1', 'Yes'), ('0', 'No')))

builtins.register(SelectRow)
