# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
Page and top-level layout widget.
"""

import os
from os import path
from wehjit.util import dumps
from wehjit import base, builtins, staticdir
from wehjit.state import *


class Page(base.Page):
    def populate_standard(self, add_view=True):
        self.topbar = self.new_child('TopBar', id='topbar')
        self.menuset = self.topbar.new_child('MenuSet', id='menuset')
        self.menu = self.menuset.new_child('Menu', id='menu')
        self.actions = self.topbar.new_child('Actions', id='actions')
        if add_view:
            self.view = self.new_child('View', id='view')

    style_vars = (
        ('color', '#000'),
        ('color_bar', '#fff'),
        ('color_error', '#d02'),

        ('colorbg', '#fff'),
        ('colorbg_code', '#eee'),
        ('colorbg_dialog_message', '#eee'),
        ('colorbg_bar_top', '#5b5b7e'),
        ('colorbg_bar', '#446'),
        ('colorbg_hover_menu', '#669'),
        ('colorbg_target_menu', '#966'),

        ('colorborder_menu', '#000'),
        ('colorborder_dialog_summary', '#595'),
        ('colorborder_dialog_error', '#b55'),

        ('font_family', 'sans-serif'),
        ('font_size', '11pt'),
        ('font_size_mono', '10pt'),

        ('height_bar', '2em'),

        ('padding_menu', '0 0.5em'),
        ('padding_input', '2px'),
        ('width_content', '50em'),
    )

    style_global = (
        ('body', (
            ('font-size', '%(font_size)s'),
            ('font-family', '%(font_family)s'),
            ('color', '%(color)s'),
            ('background-color', '%(colorbg)s'),
        )),

        ('input, select, option, textarea, button', (
            ('font-family', 'monospace'),
            ('font-size', '%(font_size_mono)s'),
            ('line-height', '%(height_bar)s'),
        )),

        ('code', (
            ('background-color', '%(colorbg_code)s'),
            ('padding', '1px'),
        )),

        ('code, pre', (
            ('font-family', 'monospace'),
            ('font-size', '%(font_size_mono)s'),
        )),

        ('label.associate', (
            ('-moz-appearance', 'button'),
            ('padding', '0.1em 0.2em'),
            ('display', 'inline-block'),
        )),

        ('.numeric', (
            ('text-align', 'right'),
        )),

        ('.hide', (
            ('display', 'none !important'),
        )),
    )

    javascript = """

    Element.implement({

        connect: function(event_name, instance,  meth_name){
            var meth_name = meth_name || 'on_' + event_name;
            var callback = instance[meth_name].bind(instance);
            this.addEvent(event_name, function(e) {
                console.log('Event:', event_name, this, meth_name);
                callback(this, new Event(e));
            });
        },

        set_height: function(spec) {
            if ($type(spec) == 'number') {
                spec += 'px';
            }
            this.setStyle('height', spec);
            return this;
        },

        set_width: function(spec) {
            if ($type(spec) == 'number') {
                spec += 'px';
            }
            this.setStyle('width', spec);
            return this;
        },

        set_text: function(value, sep) {
            var sep = (sep == undefined) ? ', ' : sep;
            if (value == undefined) {
                this.textContent = '';
                return;
            }
            if ($type(value) == 'array') {
                var value = value.join(sep);
            }
            this.textContent = value;
        },

        hide: function() {
            console.log('hide', this);
            this.addClass('hide');
        },

        show: function() {
            console.log('show', this);
            this.removeClass('hide');
        },

        get_data: function() {

            if (this.tagName == 'form') {
                var data = new Hash();
                $each(this.elements, function(el) {
                    if (! el.name) {
                        return;
                    }
                    var value = el.get_data();
                    if (value === null) {
                        return;
                    }
                    if (data.has(el.name)) {
                        var old = data.get(el.name);
                        // Replace value:
                        if (old === false) {
                            data.set(el.name, value);
                            return;
                        }
                        // Or append value:
                        if ($type(old) != 'array') {
                            var old = [old];
                            data.set(el.name, old);
                        }
                        old.push(value);
                    }
                    else {
                        data.set(el.name, value);
                    }
                });
                return data.getClean();
            }

            if (this.tagName == 'input') {
                var type = this.getAttribute('type');
                if (['text', 'password', 'hidden'].contains(type)) {
                    return this.value;
                }
                if (['checkbox', 'radio'].contains(type)) {
                    return this.checked;
                }
                return null;
            }

            if (this.tagName == 'select') {
                if (! this.multiple) {
                    return this.value;
                }
                var data = new Array();
                $each(this.options, function(opt) {
                    if (opt.selected) {
                        data.push(opt.value);
                    }
                });
                return data;
            }
        },

        set_data: function(value, sep) {
            var sep = (sep == undefined) ? ', ' : sep;
            if (this.tagName == 'form') {
                var data = new Hash(value);
                $each(this.elements, function(el) {
                    if (! el.name) {
                        return;
                    }
                    if (data.has(el.name)) {
                        el.set_data(data.get(el.name), sep);
                    }
                    else {
                        el.set_data('', sep);
                    }
                });
                return true;
            }

            if (this.tagName == 'input') {
                if ($type(value) == 'array') {
                    var value = value.join(sep);
                }
                var type = this.getAttribute('type');
                if (['text', 'password', 'hidden'].contains(type)) {
                    this.value = value;
                    return true;
                }
                if (['checkbox', 'radio'].contains(type)) {
                    this.checked = value;
                    return true;
                }
                return false;
            }

            if (this.tagName == 'select') {
                if (! this.multiple) {
                    this.value = value;
                    return true;
                }
                if ($type(value) == 'array') {
                    $each(this.options, function(opt) {
                        opt.selected = value.contains(opt.value);
                    });
                }
                else {
                    $each(this.options, function(opt) {
                        opt.selected = (opt.value == value);
                    });
                }
                return true;
            }

            return false;
        },

    });

    var Wehjit = {
        bases: {},

        plugins: {},

        instances: new Hash(),

        on_load: function() {
            console.log('Wehjit.on_load()');
            if (Wehjit.init) {
                Wehjit.init.each(Wehjit.each_plugin);
            }
            Wehjit.instances.each(function(inst) {
                if (inst.post_init) {
                    inst.post_init();
                }
            });
            Wehjit.instances.each(function(inst) {
                if (inst.post_init2) {
                    inst.post_init2();
                }
            });
        },

        each_plugin: function(plugin) {
            Wehjit.plugins[plugin.klass] = plugin;
            plugin.instances = [];
            plugin.mooclass = Wehjit.bases[plugin.uses];
            var selector = plugin.tag + '.' + plugin.klass;
            $$(selector).each(function(el) {
                Wehjit.each_instance(plugin, el);
            });
        },

        each_instance: function(plugin, el) {
            var inst = new plugin.mooclass(plugin.tag, plugin.klass, el);
            plugin.instances.push(inst);
            Wehjit.instances.set(inst.id, inst);
        },

        get: function(id) {
            return Wehjit.instances.get(id);
        },

        show: function(id) {
            var el = $(id);
            if (el.__screen__) {
                el.__screen__.show(id);
            }
            else {
                el.show();
            }
        },

    };

    window.addEvent('load', Wehjit.on_load);

    Wehjit.bases.Widget = new Class({
        Implements: Events,

        initialize: function(tag, klass, el) {
            console.log('Widget.initialize(%s, %s, %s)', tag, klass, el);
            this.tag = tag;
            this.klass= klass;
            this.el = $(el);
            this.id = this.el.id;
            console.assert($type(this.el) == 'element');
            console.assert(this.tag == this.el.tagName);
            console.assert(this.el.hasClass(this.klass));
            this.data = Wehjit.data[this.id];
            if (this.add_all_events) {
                this.add_all_events();
            }
        },

        add_event: function(event) {
            console.log('Widget.add_event(%s)', event);
            var callback = this['on_' + event];
            if (callback && this.target) {
                console.log('    Adding %s event on %s', event, this.id);
                console.log('    target = %s', this.target);
                this.target.addEvent(event, callback.bind(this));
            }
        },

        hide: function() {
            this.el.hide();
            this.fireEvent('hide', this);
        },

        show: function() {
            this.el.show();
            this.fireEvent('show', this);
        },

        get_parent: function() {
            return Wehjit.get(this.data.parent);
        },

        get_children: function() {
            var children = new Array();
            if (this.data.children) {
                this.data.children.each(function(id) {
                    console.log('Child: %s', id);
                    children.push(Wehjit.get(id));
                });
            }
            return children;
        },

    });


    Wehjit.bases.FormField = new Class({
        Extends: Wehjit.bases.Widget,

        add_all_events: function() {
            console.log('FormField.add_all_events(), id=%s', this.id);
            var fields = this.el.getElements('input, button, select, textarea');
            if (fields.length == 1) {
                this.field = fields[0];
                this.target = this.field;
            }
            else if (fields.length > 1) {
                this.fields = fields;
                this.target = this.fields;
            }
            ['change', 'focus', 'blur'].each(this.add_event, this);
        },
    });
    """

    @StaticProp
    def inline_js(self):
        return '\nWehjit.init = %s;\nWehjit.data = %s;\n' % (
            dumps(list(self._iter_js_classes())),
            dumps(dict(self._iter_json())),
        )

    require_js = ('mootools-core-nc.js',)

    @classmethod
    def register_assets(cls, assets):
        for name in os.listdir(staticdir):
            f = path.join(staticdir, name)
            if path.isfile(f):
                assets.register(f)

builtins.register(Page)


class PageApp(Page):

    def populate(self):
        self.populate_standard()
        self.h1 = self.new('H1', id='h1')
        self.view.add(self.h1)

builtins.register(PageApp)


class PageGrid(PageApp):

    def populate(self):
        self.populate_standard(add_view=False)
        self.cruds = self.actions.new_child('CRUDS', id='cruds')
        self.dialogs = self.new_child('DialogSet')
        for name in ('create', 'retrieve', 'update', 'delete', 'result'):
            child = self.dialogs.new_child('Dialog', id=name)
            setattr(self, name, child)

        self.retrieve.jumps = (
            ('update', 'Update'),
            ('delete', 'Delete'),
        )

        self.create.action = 'create'
        self.create.action_label = 'Create'

        self.update.action = 'update'
        self.update.action_label = 'Update'

        self.delete.action = 'delete'
        self.delete.action_label = 'Delete'

        form = self.create.new_child('Form')
        self.create.fieldtable = form.new_child('FieldTable')

        form = self.update.new_child('Form')
        self.update.fieldtable = form.new_child('FieldTable')

        self.retrieve.display = self.retrieve.new_child('Display')
        self.delete.display = self.delete.new_child('Display')
        self.result.display = self.result.new_child('Display')

        self.grid = self.new_child('Grid', id='grid')

builtins.register(PageGrid)


class H1(base.Widget):
    """
    An <h1> element with the page title.
    """

    label = Dynamic('label')

    @DynamicProp
    def page_title(self):
        if self.page is None:
            return
        return self.page.title

    xml = """
    <h1
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
        py:content="label or page_title"
    />
    """

    style = (
        ('', (
            ('font-size', '1.5em'),
            ('margin', '0.25em 0'),
        )),
    )

builtins.register(H1)


class TopBar(base.Container):
    """
    A fixed-position div/bar at top of page.
    """

    style = (
        ('', (
            ('position', 'fixed'),
            ('height', '%(height_bar)s'),
            ('top', '0'),
            ('left', '0'),
            ('right', '0'),
            ('z-index', '50'),
            ('line-height', '%(height_bar)s'),
            ('color', '%(color_bar)s'),
            ('background-color', '%(colorbg_bar)s'),
            (
                'background',
                '-moz-linear-gradient(top, %(colorbg_bar_top)s, %(colorbg_bar)s)'
            ),
        )),
    )

builtins.register(TopBar)


class Actions(base.Container):

    style = (

        ('', (
            ('text-align', 'right'),
            ('height', '100%%'),
        )),

        ('label', (
            ('margin', '0 0.25em'),
        )),

        ('button', (
            ('margin', '0 0.25em'),
        )),

        ('input', (
            ('border', 'none'),
            ('padding', '%(padding_input)s'),
            ('margin', '0 0.25em'),
        )),
    )

builtins.register(Actions)


class DialogSet(base.Container):
    """
    Contains one or more `Dialog` instances but only shows one at a time.
    """

    _id = 'dialogs'

    xml = """
    <div
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
        py:if="children"
    >
    <div class="overlay hide" />
    <child py:for="child in children" py:replace="child.generate()" />
    </div>
    """

    style = (
        ('div.overlay', (
            ('position', 'fixed'),
            ('top', '0'),
            ('right', '0'),
            ('bottom', '0'),
            ('left', '0'),
            ('z-index', '100'),
            ('background-color', '#000'),
            ('opacity', '0.5'),
        )),
    )

    js_class = 'DialogSet'

    javascript = """
    Wehjit.bases.DialogSet = new Class({
        Extends: Wehjit.bases.Widget,

        post_init: function() {
            this.overlay = this.el.getElement('div.overlay');
            console.assert(this.overlay);
            this.dialogs = new Hash();
            this.data.children.each(function(id) {
                var w = Wehjit.get(id);
                if (w && w.klass == 'Dialog') {
                    this.dialogs.set(id, w);
                    w.addEvent('close', this.on_close.bind(this));
                }
            }, this);
            this.current = null;
        },

        on_close: function(dialog, params) {
            this.current.el.hide();
            this.overlay.hide();
            this.current = null;
        },

        get: function(id) {
            return this.dialogs.get(id);
        },

        close: function() {
            if (this.current) {
                this.current.close();
            }
        },

        run: function(id, params) {
            if (this.current) {
                this.current.el.hide();
            }
            this.current = this.get(id);
            this.overlay.show();
            this.current.el.show();
            this.current.do_run(params);
        },
    });
    """

builtins.register(DialogSet)


class Dialog(base.Container):
    close_label = Static('close_label', default='Close')
    cancel_label = Static('cancel_label', default='Cancel')
    action_label = Static('action_label', default='Submit')

    action = Static('action')
    jumps = Static('jumps', default=tuple())

    title = Dynamic('title', json=True)
    message = Dynamic('message', doc='Message text to display')
    error = Dynamic('error', default=False, doc='If True, message is an error')

    xml = """
    <div
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes} hide"
        id="${id}"
    >
    <div class="bar">

    <div class="left">
    <button
        py:for="(name, label) in jumps"
        py:content="label"
        name="${name}"
    />
    <button py:if="action" py:content="cancel_label" name="close" />
    </div>

    <div class="center" py:content="title" />

    <div class="right">
    <button py:if="action" py:content="action_label" name="${action}" />
    <button py:if="not action" py:content="close_label" name="close" />
    </div>

    </div>

    <div class="view">
    <div
        class="${['message hide', 'message'][bool(message)]}"
    >
    <p
        class="${[None, 'error'][error]}"
        py:content="message"
    />
    </div>
    <child py:for="child in children" py:replace="child.generate()" />
    </div>

    </div>
    """

    style = (
        ('', (
            ('position', 'fixed'),
            ('top', '1em'),
            ('right', '1em'),
            ('bottom', '1em'),
            ('left', '1em'),
            ('z-index', '100'),
            ('background-color', '#fff'),
            #('-moz-box-shadow', 'black 0.5em 0.5em 0.5em'),
        )),

        ('div.bar', (
            ('position', 'absolute'),
            ('height', '%(height_bar)s'),
            ('top', '0'),
            ('left', '0'),
            ('right', '0'),
            ('line-height', '%(height_bar)s'),
            ('color', '%(color_bar)s'),
            ('background', '%(colorbg_bar)s'),
        )),

        ('div.view', (
            ('position', 'absolute'),
            ('top', '%(height_bar)s'),
            ('bottom', '0'),
            ('left', '0'),
            ('right', '0'),
            ('text-align', 'center'),
            ('overflow', 'auto'),
        )),

        ('div.bar label', (
            ('margin', '0 0.5em'),
        )),

        ('div.bar button', (
            ('margin', '0 0.5em'),
        )),

        ('div.left', (
            ('float', 'left'),
            ('width', '33%%'),
        )),

        ('div.center', (
            ('float', 'left'),
            ('width', '33%%'),
            ('text-align', 'center'),
            ('font-weight', 'bold'),
        )),

        ('div.right', (
            ('text-align', 'right'),
        )),

        ('div.message p', (
            ('background-color', '%(colorbg_dialog_message)s'),
            ('border-color', '%(colorborder_dialog_summary)s'),

            ('font-weight', 'bold'),
            ('-moz-border-radius', '100%%'),
            ('border-style', 'solid'),
            ('border-width', '2px'),
            ('padding', '0.5em'),
            ('margin', '1em 2em'),
            ('text-align', 'center'),
        )),

        ('div.message p.error', (
            ('border-color', '%(colorborder_dialog_error)s'),
            ('color', '%(color_error)s'),
        ))
    )

    js_class = 'Dialog'
    javascript = """
    Wehjit.bases.Dialog = new Class({
        Extends: Wehjit.bases.Widget,

        post_init: function() {
            this.params = null;
            this.bar = this.el.getElement('div.bar');
            console.assert(this.bar);
            this.bar.getElements('button').each(function(b) {
                b.connect('click', this);
            }, this);
            this.title = this.bar.getElement('div.center');
            console.assert(this.title);
            this.message = this.el.getElement('div.message');
            this.message_p = this.message.getElement('p');
            console.assert(this.message_p);
            this.addEvent('close', this.on_close.bind(this));
            this.dialogs = this.get_parent();
            console.assert(this.dialogs && this.dialogs.klass == 'DialogSet');
        },

        set_title: function(text) {
            this.title.textContent = text;
        },

        format_title: function(obj) {
            console.log('format title', obj);
            this.set_title(this.data.title.substitute(obj));
        },

        set_summary: function(text) {
            if (! text) {
                this.message.hide();
                return;
            }
            this.message.show();
            this.message_p.removeClass('error');
            this.message_p.textContent = text;
        },

        set_error: function(text) {
            if (! text) {
                this.message.hide();
                return;
            }
            this.message.show();
            this.message_p.addClass('error');
            this.message_p.textContent = text;
            console.log(text);
        },

        on_click: function(button, event) {
            var name = button.getAttribute('name');
            console.log('on_click: %s', name);
            this.fireEvent(name, [this, this.params]);
        },

        on_close: function(dialog, params) {
            this.params = null;
        },

        close: function() {
            this.fireEvent('close', [this, this.params]);
        },

        do_run: function(params) {
            this.params = params;
            this.fireEvent('run', [this, params]);
        },

        run: function(params) {
            this.dialogs.run(this.id, params);
        },
    });
    """


builtins.register(Dialog)



class Screen(base.Container):
    """
    Contains one or more `View` instances but only shows one at a time.
    """

    js_class = 'Screen'

    javascript = """
    Wehjit.bases.Screen = new Class({
        Extends: Wehjit.bases.Widget,

        post_init: function() {
            if (! this.data.default && this.data.children.length > 0) {
                this.data.default = this.data.children[0];
            }
            this.views = new Hash();
            this.data.children.each(function(id) {
                var el = $(id);
                el.__screen__ = this;
                this.views.set(id, el);
                if (id != this.data.default) {
                    el.hide();
                }
            }, this);
        },

        post_init2: function() {
            if (this.data.default) {
                this.show(this.data.default);
            }
        },

        show: function(id) {
            if (this.current) {
                this.current.hide();
            }
            this.current = this.views.get(id);
            this.current.show();
            this.fireEvent('show', [this, id]);
        },
    });
    """

builtins.register(Screen)


class ScreenJump(base.Widget):
    """
    A select-dropdown to switch between `Veiw` instances in a `Screen`.
    """

    screen = Static('screen', json=True)

    label = Static('label', default='View:')

    @DynamicProp
    def options(self):
        if self.screen is None:
            return tuple()
        return tuple(
            (view.id, view.title) for view in self.screen.iter()
        )

    xml = """
    <label
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
    >
    ${label}
    <select>
    <option
        py:for="(optvalue, optlabel) in options"
        value="${optvalue}"
        py:content="optlabel"
    />
    </select>
    </label>
    """

    js_class = 'ScreenJump'
    javascript = """
    Wehjit.bases.ScreenJump = new Class({
        Extends: Wehjit.bases.FormField,

        post_init: function() {
            this.screen = Wehjit.get(this.data.screen);
            console.assert(this.screen);
            this.screen.addEvent('show', this.on_show.bind(this));
        },

        on_show: function(screen, id) {
            console.log('on_show', screen, id);
            this.field.value = id;
        },

        on_change: function() {
            this.screen.show(this.field.value);
        },
    });
    """

builtins.register(ScreenJump)



class View(base.Container):
    """
    Scrollable viewport beneath a `TopBar`.
    """

    title = Dynamic('title')

    style = (
        ('', (
            ('padding', '0 1em'),
            ('overflow', 'auto'),
            ('position', 'fixed'),
            ('top', '%(height_bar)s'),
            ('right', '0'),
            ('bottom', '0'),
            ('left', '0'),
        )),
    )

builtins.register(View)



# MenuSet, Menu, and MenuItem share some common style:
ul_style = (
    ('list-style-type', 'none'),
    ('margin', '0'),
    ('padding', '0'),
)

li_style = (
    ('padding', '0'),
    ('margin', '0'),
    ('cursor', 'default'),
    ('color', '%(color_bar)s'),
)

li_cell_style = (
    ('padding', '%(padding_menu)s'),
    ('line-height', '%(height_bar)s'),
    ('color', '%(color_bar)s'),
    ('white-space', 'nowrap'),
    ('text-decoration', 'none'),
    ('display', 'block'),
)


class MenuSet(base.Container):
    """
    A group of menus, usually first child in a `TopBar`.
    """

    base_id_from = None

    xml = """
    <ul
        xmlns:py="http://genshi.edgewall.org/"
        id="${id}"
        class="${css_classes}"
    >
    <li py:for="child in children" py:if="child" py:replace="child.generate()" />
    </ul>
    """

    style = (
        ('', ul_style + (
            ('float', 'left'),
        )),

        ('> li', (
            ('float', 'left'),
        )),
    )

builtins.register(MenuSet)


class Menu(base.Container):
    """
    A menu within a `MenuSet`.
    """

    label = Dynamic('label', default='Menu Label')
    title = Dynamic('title', default='Click to open menu')

    xml = """
    <li
        xmlns:py="http://genshi.edgewall.org/"
        id="${id}"
        class="${css_classes}"
        title="${title}"
    >
    <span py:content="label" />
    <ul class="hide">
    <li py:for="child in children" py:replace="child.generate()" />
    </ul>
    </li>
    """

    style = (
        ('', li_style),

        (':hover', (
            ('background-color', '%(colorbg_hover_menu)s'),
        )),

        ('> span', li_cell_style),

        ('> ul', ul_style + (
            ('position', 'fixed'),
            ('z-index', '60'),
            ('border-width', '1px'),
            ('border-style', 'solid'),
            ('border-color', '%(colorborder_menu)s'),
            ('background-color', '%(colorbg_bar)s'),
            ('-moz-box-shadow', 'black 0.5em 0.5em 0.5em'),
        )),

#        (':hover > ul', (
#            ('display', 'block'),
#        )),
    )

    js_class = 'Menu'

    javascript = """
    Wehjit.bases.Menu = new Class({
        Extends: Wehjit.bases.Widget,

        post_init: function() {
            this.el.addEvent('click', this.on_click.bind(this));
            this.el.addEvent('mouseleave', this.on_mouseleave.bind(this));
            this.ul = this.el.getElement('ul');
            console.assert(this.ul);
        },

        on_click: function() {
            this.ul.show();
        },

        on_mouseleave: function() {
            this.ul.hide();
        },
    });
    """

builtins.register(Menu)


class MenuItem(base.Widget):
    """
    A menu-item within a `Menu`.
    """

    label = Dynamic('label', default='MenuItem Label')
    href = Dynamic('href')
    title = Dynamic('title')

    @StaticProp
    def page_url(self):
        if self.page is None:
            return
        return self.page.url

    xml = """
    <li
        xmlns:py="http://genshi.edgewall.org/"
        id="${id}"
        class="${css_classes}"
        title="${title}"
    >
    <a
        href="${relurl(href)}"
        class="${(None, 'target')[href == page_url]}"
        py:content="label"
    />
    </li>
    """

    style = (
        ('', li_style),

        ('a', li_cell_style),

        ('a.target', (
            ('background-color', '%(colorbg_target_menu)s'),
        )),

        (':hover > a', (
            ('background', '%(colorbg_hover_menu)s'),
        )),
    )

builtins.register(MenuItem)


class QuickJump(base.Widget):
    """
    A select-dropdown to move between anchors in a page.
    """

    label = Dynamic('label', default='Jump:')
    options = Dynamic('options', default=tuple())

    xml = """
    <label
        xmlns:py="http://genshi.edgewall.org/"
        class="${css_classes}"
        id="${id}"
    >
    ${label}
    <select name="${name}">
    <option
        py:for="(optvalue, optlabel) in options"
        value="#${optvalue}"
        py:content="optlabel"
    />
    </select>
    </label>
    """

    js_class = 'QuickJump'
    javascript = """
    Wehjit.bases.QuickJump = new Class({
        Extends: Wehjit.bases.FormField,

        post_init: function() {
            this.start();
        },

        start: function() {
            if (this.timerid) {
                this.stop();
                console.assert(! this.timerid);
            }
            this.timerid = this.poll.periodical(250, this);
        },

        stop: function() {
            this.timerid = $clear(this.timerid);
        },

        poll: function() {
            if (this.field.value != window.location.hash) {
                this.field.value = window.location.hash;
            }
        },

        on_change: function() {
            window.location.hash = this.field.value;
        },

        on_focus: function() {
            this.stop();
        },

        on_blur: function() {
            this.start();
        },

    });
    """

builtins.register(QuickJump)
