# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# wehjit: A Python web-widget library
# Copyright (C) 2009  Red Hat
#
# This file is part of `wehjit`.
#
# `wehjit` is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# `wehjit` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# `wehjit`.  If not, see <http://www.gnu.org/licenses/>.

"""
HTTP, WSGI, and misc. utility functions and classes.
"""

from compat import parse_qs, json
from genshi.template import MarkupTemplate


def serialize(obj):
    if callable(getattr(obj, '__json__', None)):
        return obj.__json__()
    raise TypeError('%r is not JSON serializable' % obj)


def dumps(obj):
    return json.dumps(obj, sort_keys=True, indent=4, default=serialize)


class Alternator(object):
    """
    Alternate between multiple CCS classes.

    This provides a generalized solution to the problem of highlighting every
    other row in a table.  For example:

    >>> alt = Alternator(['odd', 'even'])
    >>> alt.next()  # Row 1
    'odd'
    >>> alt.next()  # Row 2
    'even'
    >>> alt.next()  # Row 3
    'odd'
    >>> alt.next()  # Row 4
    'even'

    However, you can provide any number of classes through which to alternate:

    >>> three_way = Alternator(['red', 'green', 'blue'])
    >>> list(three_way.next() for i in xrange(6))
    ['red', 'green', 'blue', 'red', 'green', 'blue']

    Lastly, the optional ``step`` kwarg, which by default is ``1``, can be used
    to change class only every ``step`` rows.

    >>> group_three = Alternator(['odd', 'even'], step=3)
    >>> list(group_three.next() for i in xrange(6))
    ['odd', 'odd', 'odd', 'even', 'even', 'even']
    """

    def __init__(self, classes, step=1):
        self.__classes = tuple(classes)
        self.__step = step
        self.reset()

    def __repr__(self):
        return '%s(%r, step=%r)' % (
            self.__class__.__name__, self.__classes, self.__step
        )

    def next(self):
        if not self.__classes:
            return None
        i = (self.__counter / self.__step) % len(self.__classes)
        self.__counter += 1
        return self.__classes[i]

    def reset(self):
        self.__counter = 0


def read_input(environ):
    """
    Read the request body from environ['wsgi.input'].
    """
    try:
        length = int(environ.get('CONTENT_LENGTH'))
    except (ValueError, TypeError):
        return
    return environ['wsgi.input'].read(length)


def ispost(environ):
    """
    Return ``True`` if environ is a "POST" request.
    """
    return environ['REQUEST_METHOD'] == 'POST'


def isget(environ):
    """
    Return ``True`` if environ is a "GET" request.
    """
    return environ['REQUEST_METHOD'] == 'GET'


def isform(environ):
    """
    Return ``True`` if the Content-Type is "application/x-www-form-urlencoded".
    """
    return environ['CONTENT_TYPE'] == 'application/x-www-form-urlencoded'


def nicify_query(query, encoding='utf-8'):
    if not query:
        return
    for (key, value) in query.iteritems():
        if len(value) == 0:
            yield (key, None)
        elif len(value) == 1:
            yield (key, value[0].decode(encoding))
        else:
            yield (key, tuple(v.decode(encoding) for v in value))


def extract_query(environ):
    """
    Return the query as a ``dict``, or ``None`` if no query is presest.
    """
    qstr = None
    if environ['REQUEST_METHOD'] == 'POST':
        if environ['CONTENT_TYPE'] == 'application/x-www-form-urlencoded':
            qstr = read_input(environ)
    elif environ['REQUEST_METHOD'] == 'GET':
        qstr = environ['QUERY_STRING']
    if qstr:
        query = dict(nicify_query(
            parse_qs(qstr, keep_blank_values=True)
        ))
    else:
        query = {}
    environ['wsgi.query'] = query
    return query


_not_found_template = MarkupTemplate(
"""<html xmlns:py="http://genshi.edgewall.org/">
<head>
<title>404 Not Found</title>
</head>
<body>
<h1>Not Found</h1>
<p>
The requested URL <a href="${uri}" py:content="uri" /> was not found on this server.
</p>
</body>
</html>""")


def not_found(environ, start_response):
    """
    Return a 404 Not Found error.
    """
    status = '404 Not Found'
    response_headers = [('Content-Type', 'text/html')]
    start_response(status, response_headers)
    return [
        _not_found_template.generate(
            uri=(environ['SCRIPT_NAME'] + environ['PATH_INFO']),
        ).render('html')
    ]
